// Ben Clewett, April 2004
//
// Wrapper for SQLite to return a query as a table in memory.
//
// There is no warrenty, the code provided for illistation reasons only. Any user withing to make use of this does so at their own risk.


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "lite_wrap.h"

#define MAX_GET_LOCK_ATTEMPT 10

int lw_error = SQLITE_OK;
const char *lw_errmsg;

unsigned char data_empty[1] = { 0 };
char name_unknown[] = "?";

// Internal functions:
lw_result *empty_table();
lw_result *scalar_table(unsigned char *data);
lw_result *malloc_result();
lw_cell_data *malloc_cell();
lw_meta_data *malloc_meta_data(int count);
const unsigned char *lw_u_strdup(const unsigned char *data);
const unsigned char *lw_cell_print(lw_result *result, int col, int row);
void lw_divider(lw_result *result, FILE *f, int *col_size);
void lw_show_escape(FILE *f, char *field);
int lw_getlength(const unsigned char *field);
int lw_get_row_height(lw_result *result, int row);
void lw_show_line(FILE *f, const unsigned char *s, int line, int length);

lw_result *lw_query(sqlite3 *handle, const char *query)
{
	static lw_result *result;
	static lw_cell_data *cell, *last_cell;
	static sqlite3_stmt *h_statement;
	static const char *tail;
	static const unsigned char *data;
	static int current_col, i, return_value;

	h_statement = NULL;

	if (!handle) {
		lw_error = SQLITE_ERROR;
		lw_errmsg = "Null Handle";
		return NULL;
	}

	if (!query) {
		lw_error = SQLITE_ERROR;
		lw_errmsg = "Null Query";
		return NULL;
	}

	if ((return_value = sqlite3_prepare(handle, query, -1, &h_statement, &tail)) != SQLITE_OK) {
		lw_error = return_value;
		lw_errmsg = sqlite3_errmsg(handle);
		return NULL;
	}

	result = malloc_result();
	if (!result) {
		lw_error = SQLITE_NOMEM;
		lw_errmsg = "Out of Memory";
		return NULL;
	}

	result->row_count = 0;
	last_cell = NULL;

	while (1) {

		int lock_count = 0;

		do {
			return_value = sqlite3_step(h_statement);
		} while ((return_value == SQLITE_BUSY || return_value == SQLITE_LOCKED) && lock_count < MAX_GET_LOCK_ATTEMPT);

		if (return_value != SQLITE_ROW) break;

		result->col_count = sqlite3_column_count(h_statement);

		if (result->row_count == 0) {
			result->meta_data = malloc_meta_data(result->col_count);
			if (!result->meta_data) {
				lw_free_result(result);
				lw_error = SQLITE_NOMEM;
				lw_errmsg = "Out of Memory";
				return NULL;
			}
			for (current_col = 0; current_col < result->col_count; current_col++) {
				result->meta_data[current_col].type = sqlite3_column_type(h_statement, current_col);
				result->meta_data[current_col].name = strdup(sqlite3_column_name(h_statement, current_col));
				if (!result->meta_data[current_col].name) {
					result->meta_data[current_col].name = name_unknown;
				}
			}

		}

		for (current_col = 0; current_col < result->col_count; current_col++) {

			cell = malloc_cell();
			if (!cell) {
				lw_free_result(result);
				lw_error = SQLITE_NOMEM;
				lw_errmsg = "Out of Memory";
				return NULL;
			}

			if (last_cell) last_cell->next = cell;
			last_cell = cell;

			if (!result->cell_list) result->cell_list = cell;

			data = sqlite3_column_text(h_statement, current_col);

			if (data) {
				cell->data = lw_u_strdup(data);
				if (!cell->data) {
					lw_free_result(result);
					lw_error = SQLITE_NOMEM;
					lw_errmsg = "Out of Memory";
					return NULL;
				}
			} else {
				cell->data = NULL;
			}
		}

		result->row_count++;
	}

	if (return_value == SQLITE_BUSY || return_value == SQLITE_LOCKED) {
		lw_free_result(result);
		lw_error = return_value;
		lw_errmsg = "Cannot obtain lock or database busy.";
		return NULL;
	}

	if (return_value != SQLITE_DONE) {
		lw_free_result(result);
		lw_error = return_value;
		lw_errmsg = sqlite3_errmsg(handle);
		return NULL;
	}
	if (!result->row_count) {
		lw_free_result(result);
		sqlite3_finalize(h_statement);
		return empty_table();
	}

	result->cell_array = (lw_cell_data **)malloc(sizeof(lw_cell_data *) * result->row_count * result->col_count);
	if (!result->cell_array) {
		lw_free_result(result);
		lw_error = SQLITE_NOMEM;
		lw_errmsg = "Out of Memory";
		return NULL;
	}

	cell = result->cell_list;
	for (i = 0; i < result->row_count * result->col_count; i++) {
		result->cell_array[i] = cell;
		cell = (lw_cell_data *)cell->next;
	}

	sqlite3_finalize(h_statement);

	lw_error = SQLITE_OK;
	lw_errmsg = "";

	return result;
}

lw_result *empty_table()
{
	static lw_result *result;

	result = malloc_result();
	if (!result) {
		lw_error = SQLITE_NOMEM;
		lw_errmsg = "Out of Memory";
		return NULL;
	}
	
	result->col_count = result->row_count = 0;
	result->cell_array = NULL;
	result->cell_list = NULL;
	result->meta_data = NULL;

	lw_error = SQLITE_OK;
	lw_errmsg = "";

	return result;

}


int lw_get_errno()
{
	return lw_error;
}

const char *lw_get_errmsg()
{
	return lw_errmsg;
}

void lw_free_result(lw_result *result)
{
	lw_cell_data *cell, *prev_cell;
	int i;

	if (!result) return;

	if (result->meta_data) {
		for (i = 0; i < result->col_count; i++) {
			if (result->meta_data[i].name != name_unknown) free(result->meta_data[i].name);
		}
	}

	cell = result->cell_list;
	while (cell) {
		if (cell->data != data_empty) free((char *)cell->data);
		prev_cell = cell;
		cell = (lw_cell_data *)cell->next;
		free(prev_cell);
	}

	free(result->cell_array);

	free(result->meta_data);

	free(result);

}

unsigned char lw_is_null(lw_result *result, unsigned int row, unsigned int col)
{
	static int i;

	if (!result) return 0;

	i = (row * result->col_count) + col;

	return (result->cell_array[i]->data) ? 0 : 1;

}

const unsigned char *lw_cell(lw_result *result, unsigned int row, unsigned int col)
{
	static int i;
	static const unsigned char *data;

	if (!result) return data_empty;

	i = (row * result->col_count) + col;
	data = result->cell_array[i]->data;

	return (data) ? data : data_empty;
}

int lw_cell_i(lw_result *result, unsigned int row, unsigned int col)
{
	static int i;
	i = atoi((const char *)lw_cell(result, row, col));
	return i;
}
double lw_cell_d(lw_result *result, unsigned int row, unsigned int col)
{
	static double d;
	d = atof((const char *)lw_cell(result, row, col));
	return d;
}

// Malloc functions used so that can NULL content
// therefore can FREE data without core dump.
lw_result *malloc_result()
{
	static lw_result *result;

	result = (lw_result *)malloc(sizeof(lw_result));
	
	if (!result) return NULL;

	result->cell_array = NULL;
	result->cell_list = NULL;
	result->meta_data = NULL;
	result->row_count = 0;
	result->col_count = 0;

	return result;
}

lw_cell_data *malloc_cell()
{
	static lw_cell_data *cell;

	cell = (lw_cell_data *)malloc(sizeof(lw_cell_data));
	
	if (!cell) return NULL;

	cell->data = NULL;
	cell->next = NULL;

	return cell;
}

lw_meta_data *malloc_meta_data(int count)
{
	lw_meta_data *data;
	int i;

	data = (lw_meta_data *)malloc(sizeof(lw_meta_data) * count);
	if (!data) return NULL;

	for (i = 0; i < count; i++) {
		data[i].type = 0;
		data[i].name = NULL;
	}

	return data;

}


unsigned int lw_col_count(lw_result *result)
{
	if (!result) return 0;

	return result->col_count;
}

unsigned int lw_row_count(lw_result *result)
{
	if (!result) return 0;

	return result->row_count;

}

const unsigned char *lw_u_strdup(const unsigned char *data)
{
	static unsigned char *copy_data;
	static int i, len;

	for (i = 0; data[i]; i++) ;
	len = i;

	copy_data = (unsigned char *)malloc(sizeof(unsigned char) * (len + 1));

	for (i = 0; i < len; i++) copy_data[i] = data[i];
	copy_data[len] = 0;

	return copy_data;
}

void lw_show_table(lw_result *result, FILE *f)
{
	static int col_size[1024], col, row, len;
	static char format[20];
	static int line;
	static int row_height;

	if (lw_row_count(result) == 0) return;

	// Get the colunm size
	for (col = 0; col < lw_col_count(result); col++) {
		col_size[col] = 0;
		for (row = 0; row < lw_row_count(result); row++) {
			len = lw_getlength(lw_cell_print(result, col, row));
			if (len > col_size[col]) col_size[col] = len;
		}
		len = strlen(result->meta_data[col].name);
		if (len > col_size[col]) col_size[col] = len;
	}

	lw_divider(result, f, col_size);

	// Header:
	fprintf(f, "|");
	for (col = 0; col < lw_col_count(result); col++) {
		sprintf(format, "%%-%ds|", col_size[col]);
		fprintf(f, format, result->meta_data[col].name);
	}
	fprintf(f, "\n");
	fflush(f);

	lw_divider(result, f, col_size);

	// Data:
	for (row = 0; row < lw_row_count(result); row++) {

		row_height = lw_get_row_height(result, row);

		for (line = 0; line < row_height; line++) {
			fprintf(f, "|");
			for (col = 0; col < lw_col_count(result); col++) {
				lw_show_line(f, lw_cell_print(result, col, row), line, col_size[col]);
			}
			fprintf(f, "\n");
		}

	}

	lw_divider(result, f, col_size);

}

void lw_show_line(FILE *f, const unsigned char *s, int line, int length)
{
	static unsigned char *s0, *s1;
	static int i, cr_count;

	s0 = s1 = 0;

	if (line == 0) s0 = (unsigned char *)s;

	cr_count = 0;
	for (i = 0; s[i]; i++) {
		if (s[i] == '\n') {
			cr_count++;
			if (cr_count == line) s0 = (unsigned char *)s + 1 + i;
			if (cr_count == line + 1) s1 = (unsigned char *)s + i;
		}
	}

	if (!s1) s1 = (unsigned char *)s + i;
	if (!s0) s0 = s1 = (unsigned char *)s;

	for (i = (int)s0; i < (int)s1; i++) fprintf(f, "%c", *(char *)i);
	for (;i < (int)s0 + length; i++) fprintf(f, " ");

	fprintf(f, "|");
	fflush(f);

}

void lw_divider(lw_result *result, FILE *f, int *col_size)
{
	static int i, col;

	fprintf(f, "+");
	for (col = 0; col < lw_col_count(result); col++) {
		for (i = 0; i < col_size[col]; i++) fprintf(f, "-");
		fprintf(f, "+");
	}
	fprintf(f, "\n");
	fflush(f);
}

int lw_getlength(const unsigned char *field)
{
	static int max_len, test_len, i;
	max_len = 0, test_len = 0;

	for (i = 0; field[i]; i++) {
		if (field[i] == '\n' || field[i] == '\r') {
			if (test_len > max_len) max_len = test_len;
			test_len = 0;
		} else {
			test_len++;
		}
	}

	return (test_len > max_len) ? test_len : max_len;

}

int lw_get_row_height(lw_result *result, int row)
{
	static int col;
	static int max_height, test_height, i;
	static unsigned char *cell;

	max_height = 0;

	for (col = 0; col < lw_col_count(result); col++) {
		cell = (unsigned char *)lw_cell_print(result, col, row);
		test_height = 1;
		for (i = 0; cell[i]; i++) {
			if (cell[i] == '\n') test_height++;
		}
		if (test_height > max_height) max_height = test_height;
	}

	return max_height;
}

const unsigned char *lw_cell_print(lw_result *result, int col, int row)
{
	static unsigned char *s = NULL;
	static int i, s_len_malloc = -1, s_len_in;
	static unsigned char *cell;

	// NULL
	if (lw_is_null(result, row, col)) return (const unsigned char *)"NULL";

	cell = (unsigned char *)lw_cell(result, row, col);
	s_len_in = strlen(cell);

	if (s_len_in > s_len_malloc) {
		free(s);
		s = (unsigned char *)malloc(s_len_in + 1);
		s_len_malloc = s_len_in;
	}

	for (i = 0; cell[i]; i++) {
		if (cell[i] == '\n') s[i] = '\n';
		else if (cell[i] == '\t') s[i] = ' ';
		else if (cell[i] < ' ') s[i] = '.';
		else s[i] = cell[i];
	}
	s[i] = 0;

	return s;

}

void lw_show_table_cvs(lw_result *result, FILE *f)
{
	int col, row;

	// Header:
	for (col = 0; col < lw_col_count(result); col++) {
		fprintf(f, "%s", (col == 0) ? "" : ",");
		lw_show_escape(f, result->meta_data[col].name);
	}
	fprintf(f, "\n");
	fflush(f);

	// Data:
	for (row = 0; row < lw_row_count(result); row++) {
		for (col = 0; col < lw_col_count(result); col++) {
			fprintf(f, "%s", (col == 0) ? "" : ",");
			lw_show_escape(f, (char *)lw_cell(result, row, col));
		}
		fprintf(f, "\n");
		fflush(f);
	}

}

void lw_show_table_xml(lw_result *result, FILE *f, char *table_tag_name, char *row_tag_name)
{
	int col, row;

	// Header
	printf("<%s>\n", table_tag_name);

	// Data:
	for (row = 0; row < lw_row_count(result); row++) {

		fprintf(f, "  <%s", row_tag_name);
		for (col = 0; col < lw_col_count(result); col++) {
			fprintf(f, " %s=", result->meta_data[col].name);
			lw_show_escape(f, (char *)lw_cell(result, row, col));
		}
		fprintf(f, " />\n");
		fflush(f);
	}

	printf("</%s>\n", table_tag_name);
}

void lw_show_table_html(lw_result *result, FILE *f)
{
	int col, row;

	// Header
	printf("<TABLE BORDER=1 CELLPADDING=2 CELLSPACING=0>\n");
	fprintf(f, "  <TR>\n");
	for (col = 0; col < lw_col_count(result); col++) {
		fprintf(f, "    <TD ALIGN=CENTER><B><FONT FACE=\"Courier\">%s</FONT></B></TD>\n", result->meta_data[col].name);
	}
	fprintf(f, "  </TR>\n");
	fflush(f);

	// Data:
	for (row = 0; row < lw_row_count(result); row++) {

		fprintf(f, "<TR>\n");
		for (col = 0; col < lw_col_count(result); col++) {
			fprintf(f, "    <TD><FONT FACE=\"Courier\">%s</FONT></TD>\n", lw_cell_print(result, col, row));
		}
		fprintf(f, "</TR>\n");
		fflush(f);
	}

	fprintf(f, "</TABLE>\n");
}


void lw_show_escape(FILE *f, char *field)
{
	int i;

	fprintf(f, "\"");

	for (i = 0; field[i]; i++) {
		if (field[i] == '\"') fprintf(f, "\"\"");
		else if (field[i] >= ' ') fprintf(f, "%c", field[i]);
		else fprintf(f, ".");
	}

	fprintf(f, "\"");

}



